/**
 * This program is free software, you can redistribute it and/or
 * modify it under the terms of the GNU General Public License, a copy
 * of which is provided under the name LICENSE.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Copyright 2004 Robert Osada
 * $Id: DNAFreq.cpp,v 1.55 2004/06/10 17:52:27 rosada Exp $
 **/
#include "DNA.h"
#include "DNAFreq.h"

void DNAFreq::clear () {
  memset(count, 0, sizeof(int   )*4);  total=0;
  memset(freq , 0, sizeof(double)*4);  ic=0;
  memset(score, 0, sizeof(double)*4);
}

// Clear and set gc content.
void DNAFreq::setGC (double gc) {
  assert(gc>0 && gc<1);
  clear();
  freq[0] = freq[1] = (1 - gc) / 2;
  freq[2] = freq[3] =      gc  / 2;
  updateEntropy(ET_Freq);
}

// Select a random base with a given frequency.
char DNAFreq::randBase () {
  double r = randUnit();
  double f;
  f  = freq[0]; if (r < f) return 'a';
  f += freq[1]; if (r < f) return 't';
  f += freq[2]; if (r < f) return 'g';
  return 'c';
}

// Add counts.
void DNAFreq::add (DNAFreq &a) {
  for (int m=0; m<4; m++)
    add(m, a.count[m]);
}

// Update derived quanitities.
void DNAFreq::updateFreq (FreqType ft, DNAFreqPtr bg) {
  switch (ft) {
  case FT_Naive:
    if (!total) error("zero total when calculating naive frequencies");
    for (int m=0; m<4; m++)
      freq[m] = (double)count[m] / total;
    break;
  case FT_Bg:
    if (!bg) error("null background when calculating background adjusted frequencies");
    for (int m=0; m<4; m++)
      freq[m] = (count[m] + bg->freq[m]) / (total + 1);
    break;
  case FT_Laplace:
    for (int m=0; m<4; m++)
      freq[m] = (double)(count[m] + 1) / (total + 4);
    break;
  case FT_Skip:
    break;
  default:
    error("unknown frequency type");
  }
}
void DNAFreq::updateEntropy (EntropyType et) {
  switch (et) {
  case ET_Count:
    if (!total) error("zero total when calculating count based entropy");
  case ET_Freq:
    ic = 2;
    for (int m=0; m<4; m++) {
      double f = (et == ET_Count)
	? ((double)count[m] / total)
	: freq[m];
      if ((f > 0) && (f < 1)) ic += f * log2(f);
    }
    ic = min(max(ic, 0.), 2.);
    break;
  case ET_Skip:
    break;
  default:
    error("unknown frequency type");
  }
}
void DNAFreq::updateScore (ScoreType st, DNAFreqPtr bg) {
  switch (st) {
  case ST_Freq:
  case ST_FreqIc:
    memcpy(score, freq, sizeof(double)*4);
    break;
  case ST_LogOdds:
  case ST_LogOddsIc:
    if (!bg) error("null background when computing log-odds score");
    for (int m=0; m<4; m++) {
      assert(bg->freq[m] > 0);
      score[m] = log2(freq[m] / bg->freq[m]);
    }
    break;
  case ST_BergHippel:
  case ST_BergHippelIc: {
    /**
     * Berg / von Hippel:
     *   log((n_{l,b} + 0.5)/ (n_{l,0}+0.5)) 
     * where
     *   n_{l, b} is the number of occurrences of base b at position l in
     *   the set of binding sites, n_{l,0} is the number of occurrences of the
     *   most common base at position l in the set of binding sites.
     **/
    int c = max(count[0], count[1], count[2], count[3]);
    for (int m=0; m<4; m++)
      score[m] = log2((count[m] + .5) / (c + .5));
    break;
  }
  case ST_Consensus:
  case ST_ConsensusIc: {
    /**
     * Consensus:
     *
     * Compute the probabilities of each nucleotide, say these are
     *   f_1 >= f_2 >= f_3 >= f_4,
     * then if f_1 > .5, then nucleotide 1 is the consensus,
     * otherwise, if f_1+f_2 > 0.75 then the consensus is base1/base2,
     *   (i.e.: if have either base1 or base2, then you have a match);
     * otherwise, you don't have a consensus (so no matches).    
     **/
    double f[4] = {freq[0], freq[1], freq[2], freq[3]};
    qsort(f, 4, sizeof(double), _doubleRCmp);
    double cutoff =
      (f[0]      > .5 ) ? f[0] :
      (f[0]+f[1] > .75) ? f[1] : 2;  
    for (int m=0; m<4; m++)
      score[m] = (freq[m] >= cutoff) ? 1 : 0;
    break;
  }
  case ST_Skip:
    break;
  default:
    error("unknown score type");
  }
  switch (st) {
  case ST_FreqIc:
  case ST_LogOddsIc:
  case ST_BergHippelIc:
  case ST_ConsensusIc:
    for (int m=0; m<4; m++)
      score[m] *= ic;
    break;
  default:
    break;
  }
}

void DNAFreq::expectedScore (DNAFreq &r, double &mu, double &var) {
  double x=0, x2=0;
  for (int m=0; m<4; m++) {
    double f = r.freq[m];
    double s = score [m];
    x  += f * s    ;
    x2 += f * s * s;
  }
  mu  += x;
  var += max(x2-x*x, 0.);
}

// Print lines up with DNAPair::print.
void DNAFreq::print (FILE *file, cstring prefix, cstring suffix) {
  assert(file);
  fprintf(file, "%s a=%f   t=%f   g=%f   c=%f  gc=%f  ic=%f%s",
	  prefix, freq[0], freq[1], freq[2], freq[3], getGC(), ic, suffix);
}
