/**
 * This program is free software, you can redistribute it and/or
 * modify it under the terms of the GNU General Public License, a copy
 * of which is provided under the name LICENSE.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Copyright 2004 Robert Osada
 * $Id: DNAPair.cpp,v 1.40 2004/06/10 17:53:01 rosada Exp $
 **/
#include "DNA.h"
#include "DNAPair.h"
#include "Misc.h"

void DNAPair::clear () {
  memset(count, 0, sizeof(int   )*4*4);  total=0;
  memset(freq , 0, sizeof(double)*4*4);  ic=0;  mi=0;
  memset(score, 0, sizeof(double)*4*4);
}

void DNAPair::setInd (DNAFreq &fi, DNAFreq &fj) {
  clear();
  total = fi.total * fj.total;
  for (int m=0; m<4; m++)
  for (int n=0; n<4; n++) {
    count[m][n] = fi.count[m] * fj.count[n];
    freq [m][n] = fi.freq [m] * fj.freq [n];
  }
  ic = fi.ic + fj.ic;
}

void DNAPair::add (DNAPair &a) {
  for (int m=0; m<4; m++)
  for (int n=0; n<4; n++)
    add(m,n, a.count[m][n]);
}

void DNAPair::updateFreq (FreqType ft, DNAPairPtr bg) {
  switch (ft) {
  case FT_Naive:
    if (!total) error("zero total when calculating naive frequencies");
    for (int m=0; m<4; m++)
    for (int n=0; n<4; n++)
      freq[m][n] = (double)count[m][n] / total;
    break;
  case FT_Bg:
    if (!bg) error("null background when calculating background adjusted frequencies");
    for (int m=0; m<4; m++)
    for (int n=0; n<4; n++)
      freq[m][n] = (count[m][n] + bg->freq[m][n]) / (total + 1);
    break;
  case FT_Laplace:
    for (int m=0; m<4; m++)
    for (int n=0; n<4; n++)
      freq[m][n] = (double)(count[m][n] + 1) / (total + 16);
    break;
  case FT_Skip:
    break;
  default:
    error("unknown frequency type");
  }
}
void DNAPair::updateEntropy (EntropyType et) {
  switch (et) {
  case ET_Count:
    if (!total) error("zero total when calculating count based entropy");
  case ET_Freq: {
    ic = 4;
    DNAFreq fi, fj;  // marginals
    for (int m=0; m<4; m++)
    for (int n=0; n<4; n++) {
      double f = (et == ET_Count)
	? ((double)count[m][n] / total)
	: freq[m][n];
      if ((f > 0) && (f < 1))
	ic += f * log2(f);
      fi.freq[m] += f;
      fj.freq[n] += f;
    }
    ic = min(max(ic, 0.), 4.);
    fi.updateEntropy(ET_Freq);
    fj.updateEntropy(ET_Freq);
    mi = ic - fi.ic - fj.ic;
    mi = min(max(mi, 0.), 2.);
    break;
  }
  case ET_Skip:
    break;
  default:
    error("unknown frequency type");
  }
}
void DNAPair::updateScore (ScoreType st, DNAPairPtr bg) {
  switch (st) {
  case ST_Freq:
  case ST_FreqIc:
    memcpy(score, freq, sizeof(double)*4*4);
    break;
  case ST_LogOdds:
  case ST_LogOddsIc:
    if (!bg) error("null background when computing log-odds score");
    for (int m=0; m<4; m++)
    for (int n=0; n<4; n++) {
      assert(    freq[m][n] > 0);
      assert(bg->freq[m][n] > 0);
      score[m][n] = log2(freq[m][n] / bg->freq[m][n]);
    }
    break;
  case ST_BergHippel:
  case ST_BergHippelIc: {
    // see DNAFreq for description
    int c = 0;
    for (int m=0; m<4; m++)
    for (int n=0; n<4; n++)
      c = max(c, count[m][n]);
    for (int m=0; m<4; m++)
    for (int n=0; n<4; n++)
      score[m][n] = log2((count[m][n] + .5) / (c + .5));
    break;
  }
  case ST_Consensus:
  case ST_ConsensusIc: {
    // see DNAFreq for description
    double f[16] = {
      freq[0][0], freq[0][1], freq[0][2], freq[0][3],
      freq[1][0], freq[1][1], freq[1][2], freq[1][3],
      freq[2][0], freq[2][1], freq[2][2], freq[2][3],
      freq[3][0], freq[3][1], freq[3][2], freq[3][3]
    };
    qsort(f, 16, sizeof(double), _doubleRCmp);
    double cutoff =
      (f[0]      > .5 ) ? f[0] :
      (f[0]+f[1] > .75) ? f[1] : 2;
    for (int m=0; m<4; m++)
    for (int n=0; n<4; n++)
      score[m][n] = (freq[m][n] >= cutoff) ? 1 : 0;
    break;
  }
  case ST_Skip:
    break;
  default:
    error("unknown score type");
  }
  switch (st) {
  case ST_FreqIc:
  case ST_LogOddsIc:
  case ST_BergHippelIc:
  case ST_ConsensusIc:
    for (int m=0; m<4; m++)
    for (int n=0; n<4; n++)
      score[m][n] *= ic;
    break;
  default:
    break;
  }
}

void DNAPair::expectedScore (DNAFreq &r, double &mu, double &var) {
  double x=0, x2=0;
  for (int m=0; m<4; m++)
  for (int n=0; n<4; n++) {
    double f = r.freq[m] * r.freq[n];
    double s = score [m][n];
    x  += f * s    ;
    x2 += f * s * s;
  }
  mu  += x;
  var += max(x2-x*x, 0.);
}

void DNAPair::print (FILE *file, cstring prefix, cstring suffix) {
  for (int m=0; m<4; m++) {
    // lines up with DNAFreq::print
    char b = indexBase(m);
    fprintf(file, "%s%ca=%f  %ct=%f  %cg=%f  %cc=%f%s",
	    prefix, b, freq[m][0], b, freq[m][1], b, freq[m][2], b, freq[m][3], (m==3) ? "  " : suffix);
  }
  fprintf(file, "ic=%f  mi=%f%s", ic, mi, suffix);
}
