/**
 * This program is free software, you can redistribute it and/or
 * modify it under the terms of the GNU General Public License, a copy
 * of which is provided under the name LICENSE.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Copyright 2004 Robert Osada
 * $Id: Site.cpp,v 1.46 2004/07/29 15:41:16 rosada Exp $
 **/
#include <float.h>
#include "DNA.h"
#include "Site.h"


/**
 * Site.
 **/
Site::Site (cstring desc_, cstring seq_, int loc_, bool pStrand_, int len_, double score_, double pvalue_, int rank_, double fp_) :
  desc    (desc_ ? strdup(desc_) : 0),
  seq     (seq_  ? strNew(len_ ) : 0),
  loc     (loc_    ),
  pStrand (pStrand_),
  len     (len_    ),
  score   (score_  ),
  pvalue  (pvalue_ ),
  rank    (rank_   ),
  fp      (fp_     )
{
  assert(len >= 0);

  if (seq) {
    if (pStrand)
      strncpy(seq, seq_ + loc, len);
    else {
      cstring p = seq_ + loc + len - 1;
      for (int i=0; i<len; i++)
	seq[i] = comp(*p--);
    }
    seq[len] = 0;
  }
}

Site::Site (Site &s) :
  desc    (s.desc ? strdup(s.desc) : 0),
  seq     (s.seq  ? strdup(s.seq ) : 0),
  loc     (s.loc    ),
  pStrand (s.pStrand),
  len     (s.len    ),
  score   (s.score  ),
  pvalue  (s.pvalue )
{
  warning1("copying Site, desc=%s", desc ? desc : "NA");
  assert(&s != this);
  assert(len >= 0);
}

Site::~Site () {
  strFree(desc);
  strFree(seq );
  len = 0;
}

bool Site::isRComp () {
  assert(seq && len);
  return ::isRComp(seq, len);
}

void Site::print (FILE *file, cstring prefix, cstring suffix) {
  fprintf(file, "%s", prefix);
  if (desc)
    fprintf(file, "%s\t", desc);
  fprintf(file, "%d..%d\t%s",
    loc+1, loc+len,
    (seq && len && isRComp()) ? "+/-" : pStrand ? "+" : "-");
  if (seq)
    fprintf(file, "\t%s", seq);
  fprintf(file, "\t%s", doubleStr(score));
  if (pvalue >= 0)
    fprintf(file, " %s", doubleStr(pvalue));
  if (rank >= 0)
    fprintf(file, " %d", rank);
  if (fp >= 0)
    fprintf(file, " %s", doubleStr(fp));
  fprintf(file, "%s", suffix);
}


/**
 * Site comparison.
 **/
int siteCmp (SitePtr a, SitePtr b) {
  assert(a && b);
  return firstNonZero(
    doubleCmp(a->score,    b->score),
    intCmp   (a->loc,      b->loc),
    boolRCmp (a->pStrand,  b->pStrand));
}
int siteRCmp (SitePtr a, SitePtr b) {
  assert(a && b);
  return firstNonZero(
    doubleRCmp(a->score,    b->score),
    intCmp    (a->loc,      b->loc),
    boolRCmp  (a->pStrand,  b->pStrand));
}

int siteDescCmp (SitePtr a, SitePtr b) {
  assert(a && b);
  assert(a->desc && b->desc);
  return firstNonZero(
    strcmp (a->desc, b->desc),
    siteCmp(a, b));
}
int siteDescRCmp (SitePtr a, SitePtr b) {
  assert(a && b);
  assert(a->desc && b->desc);
  return firstNonZero(
    strcmp  (a->desc, b->desc),
    siteRCmp(a, b));
}


/**
 * Best site.
 **/
void BestSite::operator () (cstring desc_, cstring seq_, int loc_, bool pStrand_, int len_, double score_, double pvalue_, int rank_, double fp_)
{
  if (score_ > score) {
    desc    = desc_;
    seq     = seq_;
    loc     = loc_;
    pStrand = pStrand_;
    len     = len_;
    score   = score_;
    pvalue  = pvalue_;
    rank    = rank_;
    fp      = fp_;
  }
  else if (score_ == score)
    numTies++;
}


/**
 * Top Sites.
 **/
void TopSites::operator () (BestSite &bs) {
  if (sites.size() < n)
    sites.push(bs.newSite());
  else if (sites.top()->score < bs.score) {
    sites.pop();
    sites.push(bs.newSite());
  }
}
 
int TopSites::getAll (Sites &s) {
  int before = s.size();
  s.reserve(before + sites.size());
  
  while (!sites.empty()) {
    s.push_back(sites.top());
    sites.pop();
  }
  return s.size() - before;
}
