/**
 * This program is free software, you can redistribute it and/or
 * modify it under the terms of the GNU General Public License, a copy
 * of which is provided under the name LICENSE.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Copyright 2004 Robert Osada
 * $Id: SitePairScorer.cpp,v 1.8 2004/06/10 17:42:02 rosada Exp $
 **/
#include "DNA.h"
#include "SitePairScorer.h"

SitePairScorer::SitePairScorer (
  FreqType ft , ScoreType  st , int len   , DNAFreq &obg,
  BgType   bt_, ScoreType pst_, int scope_
) :
  SiteScorer(ft, st, len, obg),
  bt(bt_), pst(pst_), scope(scope_), obg2(obg, obg),
  bgS(scope), pairs(scope, len1)
{
  if ((scope < 1) || (scope > len1))
    error("scope out of range");
}

void SitePairScorer::clear () {
  SiteScorer::clear();
  bg2.clear();
  for (int k=0; k<scope; k++) {
    bgS[k].clear();
    int iMax = len1 - k;
    for (int i=0; i<iMax; i++)
      pairs(k,i).clear();
  }
}

void SitePairScorer::add (cstring seq, bool pStrand, int w) {
  SiteScorer::add(seq, pStrand, w);
  int kMax = (bt == BT_All) ? len1 : scope;
  if (pStrand) {
    for (int k=0; k<kMax; k++) {
      cstring s = seq;
      cstring t = s + k + 1;
      int iMax = len1 - k;
      for (int i=0; i<iMax; i++) {
	int m = baseIndex(*s++);
	int n = baseIndex(*t++);
	switch (bt) {
	case BT_Ind:                      break;
	case BT_All:   bg2   .add(m,n,w); break;
	case BT_Scope: bgS[k].add(m,n,w); break;
	default: error("unknown background type");
	}
	if (k < scope)
	  pairs(k,i).add(m,n,w);
      }
    }
  }
  else {
    for (int k=0; k<kMax; k++) {
      cstring s = seq + len1;
      cstring t = s - k - 1;
      int iMax = len1 - k;
      for (int i=0; i<iMax; i++) {
	int m = compIndex(*s--);
	int n = compIndex(*t--);
	switch (bt) {
	case BT_Ind:                      break;
	case BT_All:   bg2   .add(m,n,w); break;
	case BT_Scope: bgS[k].add(m,n,w); break;
	default: error("unknown background type");
	}
	if (k < scope)
	  pairs(k,i).add(m,n,w);
      }
    }
  }
}

void SitePairScorer::update () {
  SiteScorer::update();
  
  switch (bt) {
  case BT_Ind:
    bg2.setInd(bg, bg);
    bg2.updateScore(pst, &obg2);
    break;
  case BT_All:
    update(bg2, obg2);
    break;
  case BT_Scope:
    for (int k=0; k<scope; k++)
      update(bgS[k], obg2);
    break;
  default:
    error("unknown background type");
  }

  for (int k=0; k<scope; k++) {
    DNAPair &bg = (bt == BT_Scope)
      ? bgS[k]
      : bg2;
    int iMax = len1 - k;
    for (int i=0; i<iMax; i++)
      update(pairs(k,i), bg);
  }
}
void SitePairScorer::update (DNAPair &p, DNAPair &bg) {
  p.updateFreq   ( ft, &bg);
  p.updateEntropy(ET_Count);
  p.updateScore  (pst, &bg);
}

double SitePairScorer::operator() (cstring seq, bool pStrand) {
  double score = SiteScorer::operator()(seq, pStrand);
  if (pStrand) {
    for (int k=0; k<scope; k++) {
      cstring s = seq;
      cstring t = s + k + 1;
      int iMax = len1 - k;
      for (int i=0; i<iMax; i++) {
	int m = baseIndex(*s++);
	int n = baseIndex(*t++);
	score += pairs(k,i).score[m][n];
      }
    }
  }
  else {
    for (int k=0; k<scope; k++) {
      cstring s = seq + len1;
      cstring t = s - k - 1;
      int iMax = len1 - k;
      for (int i=0; i<iMax; i++) {
	int m = compIndex(*s--);
	int n = compIndex(*t--);
	score += pairs(k,i).score[m][n];
      }
    }
  }
  return score;
}

void SitePairScorer::expectedScore (DNAFreq &r, double &mu, double &var_) {
  // add means and variances
  double var = 0;
  SiteScorer::expectedScore(r, mu, var);
  for (int k=0; k<scope; k++) {
    int iMax = len1 - k;
    for (int i=0; i<iMax; i++)
      pairs(k,i).expectedScore(r, mu, var);
  }
  // calculate covariances between pairs and columns
  double cov = 0;
  for (int k=0; k<scope; k++) {
    int iMax = len1 - k;
    for (int i=0, j=k+1; i<iMax; i++, j++) { 
      DNAPair &p = pairs(k,i);
      double pMu=0, pVar=0;
      p.expectedScore(r, pMu, pVar);
      {
	DNAFreq &c = col[i];
	double cMu=0, cVar=0;
	c.expectedScore(r, cMu, cVar);
	for (int m=0; m<4; m++)
	for (int n=0; n<4; n++)
	  cov += r.freq[m] * r.freq[n] * p.score[m][n] * c.score[m];
	cov -= pMu * cMu;
      }
      {
	DNAFreq &c = col[j];
	double cMu=0, cVar=0;
	c.expectedScore(r, cMu, cVar);
	for (int m=0; m<4; m++)
	for (int n=0; n<4; n++)
	  cov += r.freq[m] * r.freq[n] * p.score[m][n] * c.score[n];
	cov -= pMu * cMu;
      }
    }
  }
  // calculate covariances between pairs and pairs
  // pairs considered in lexicographic (i,j) order
  for (int i=0; i<len1; i++)
  for (int j=i+1, k=0; (j<len) && (k<scope); j++, k++) {
    DNAPair &p = pairs(k,i);
    double pMu=0, pVar=0;
    p.expectedScore(r, pMu, pVar);    
    //   i2=i:   i-----j
    //   j2>j    i2-----j2
    for (int j2=j+1, k2=k+1; (j2<len) && (k2<scope); j2++, k2++) {
      DNAPair &p2 = pairs(k2,i);
      double p2mu=0, p2var=0;
      p2.expectedScore(r, p2mu, p2var);
      for (int m=0; m<4; m++)
      for (int n=0; n<4; n++)
      for (int o=0; o<4; o++)
	cov += r.freq[m] * r.freq[n] * r.freq[o] * p.score[m][n] * p2.score[m][o];
      cov -= pMu * p2mu;
    }
    //   i2=j:   i-----j
    //                 i2--j2
    for (int j2=j+1, k2=0; (j2<len) && (k2<scope); j2++, k2++) {
      DNAPair &p2 = pairs(k2,j);
      double p2mu=0, p2var=0;
      p2.expectedScore(r, p2mu, p2var);
      for (int m=0; m<4; m++)
      for (int n=0; n<4; n++)
      for (int o=0; o<4; o++)
	cov += r.freq[m] * r.freq[n] * r.freq[o] * p.score[m][n] * p2.score[n][o];
      cov -= pMu * p2mu;
    }
    //   j2=j:   i-----j
    //   i<i2<j    i2--j2
    for (int i2=i+1, k2=k-1; i2<j; i2++, k2--) {
      DNAPair &p2 = pairs(k2,i2);
      double p2mu=0, p2var=0;
      p2.expectedScore(r, p2mu, p2var);
      for (int m=0; m<4; m++)
      for (int n=0; n<4; n++)
      for (int o=0; o<4; o++)
	cov += r.freq[m] * r.freq[n] * r.freq[o] * p.score[m][n] * p2.score[o][n];
      cov -= pMu * p2mu;
    }
  }
  var  += 2 * cov;
  var_ += max(var, 0.);
}


/**
 * Rigorous pair profile.
 *
 * under the assumption that only adjacent pairs matter,
 * P[xyz] = P[x|yz] P[yz] = P[x|y] P[yz] = P[xy] P[yz] / P[y]
 *
 * the score is the log-odds of our sequence compared with
 * background.  for example,
 *
 * score = logOdds(xyz) = log P[xyz] / P[xyz|bg]
 *       = log (P[xy] P[yz] / P[y]) / (P[xy|bg] P[yz|bg] / P[y|bg])
 *       = log (P[xy] / P[xy|bg]) (P[yz] / P[yz|bg]) / (P[y] / P[y|bg])
 *       = logOdds(xy) + logOdds(yz) - logOdds(y)
 **/
PairProfileRigorous::PairProfileRigorous (int len, DNAFreq obj, BgType bt) :
  SitePairScorer(FT_Bg, ST_LogOdds, len, obj, bt, ST_LogOdds, 1)
{}

void PairProfileRigorous::update () {
  SitePairScorer::update();
  for (int m=0; m<4; m++) {
    col[0   ].score[m] = 0;
    col[len1].score[m] = 0;
  }
  for (int i=1; i<len1; i++) {
    DNAFreq &c = col[i];
    for (int m=0; m<4; m++)
      c.score[m] = -c.score[m];
  }
}


/**
 * Pair centroid.
 * Eliminates rounding errors.
 **/
PairCentroidScorer::PairCentroidScorer (int len, DNAFreq obg, int scope) :
  SitePairScorer(FT_Naive, ST_Freq, len, obg, BT_Ind, ST_Freq, scope)
{}

double PairCentroidScorer::operator() (cstring seq, bool pStrand) {
  double score = SitePairScorer::operator()(seq, pStrand);
  return round(score * size) / size;
}
