/**
 * Motif Search is free software, you can redistribute it and/or
 * modify it under the terms of the GNU General Public License, a copy
 * of which is provided under the name LICENSE.
 *
 * Motif Search is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Copyright 2004 Robert Osada
 * $Id: motifsearch.cpp,v 1.13 2004/07/30 16:45:43 rosada Exp $
 *
 * load a motif and search for it in a sequence.
 * report sites which pass the specified criteria.
 *
 * written for Mona Singh's computational biology lab at
 * Princeton University, 2004.
 **/
#include "motifsearch.h"


/**
 * Main.
 **/
void license () {
  printf(License);
  exit(-1);
}
void usage ()
{
  fprintf(stderr, Usage);
  exit(-1);
}
int main (int argc, cstring *argv)
{
  // defaults and options
  double gc          = .5;

  cstring method = "centroid";
  bool    ic     = false;
  cstring scope  = "0";
  cstring var    = 0;

  Criteria top;
  Criteria cutoff;
  Criteria pvalue;
  cstring  negSeqFile = 0;
  Criteria rank;
  Criteria fp;

  static struct option longopts[] = {
    { "gc"     , required_argument, NULL, 'g' },
    { "license", no_argument      , NULL, 'L' },
    { "method" , required_argument, NULL, 'm' },
    { "ic"     , no_argument      , NULL, 'i' },
    { "scope"  , required_argument, NULL, 's' },
    { "var"    , required_argument, NULL, 'v' },
    { "top"    , required_argument, NULL, 't' },
    { "cutoff" , required_argument, NULL, 'c' },
    { "pvalue" , required_argument, NULL, 'p' },
    { "neg"    , required_argument, NULL, 'n' },
    { "rank"   , required_argument, NULL, 'r' },
    { "fp"     , required_argument, NULL, 'f' }
  };
  int ch;
  while ((ch = getopt_long(argc, argv, "g:Lm:is:v:t:c:p:n:r:f:", longopts, NULL)) != -1)
    switch(ch) {
    case 'g': gc         =atof(optarg); break;
    case 'L': license();                break;
    case 'm': method     =optarg;       break;
    case 'i': ic         =true;         break;
    case 's': scope      =optarg;       break;
    case 'v': var        =optarg;       break;
    case 't': top   .set(optarg);  break;
    case 'c': cutoff.set(optarg);  break;
    case 'p': pvalue.set(optarg);  break;
    case 'n': negSeqFile = optarg; break;
    case 'r': rank  .set(optarg);  break;
    case 'f': fp    .set(optarg);  break;
    default:
      usage();
    }

  // if rank or fp are used, then negFile must be set
  if (rank.use && !negSeqFile) 
    printf("rank criteria requires negative examples\n"), usage();
  if (fp.use && !negSeqFile)
    printf("fp criteria requires negative examples\n"), usage();
  
  // set default criteria
  if (top.use + cutoff.use + pvalue.use + rank.use + fp.use == 0)
    top.set("1");
    
  // arguments
  argv += optind;
  argc -= optind;
  if ((argc != 1) && (argc != 2))
    usage();

  cstring motifFile = *argv++;
  cstring seqFile   = (argc == 1) ? 0 : *argv++;
  if (seqFile && !strcmp(seqFile,"-"))
    seqFile = 0;

  // load motif and make site scorer
  Seqs motif;
  readSeqs(motifFile, motif);
  
  int size = motif.size();
  if (!size)
    error1("unable to read %s", motifFile);
  
  int len = motif[0]->len;
  for (int i=0; i<size; i++) {
    Seq &seq = *motif[i];
    if (!seq.isDNA()) 
      error1("sequence %s isn't DNA", seq.sid);
    if (seq.len != len)
      error1("sequence %s doesn't match motif length", seq.sid);
  }

  DNAFreq obg(gc);  // overall background
  SiteScorerPtr ss = makeSiteScorer(method, scope, ic, var, len, obg);

  for (int i=0; i<size; i++) {
    Seq &seq = *motif[i];
    ss->add(seq.seq);
  }
  ss->update();

  double scoreAvg=0, scoreVar=0;
  ss->expectedScore(obg, scoreAvg, scoreVar);
  double scoreSd = sqrt(scoreVar);

  setEvdParams(scoreAvg, scoreSd, 2);
  
  // read negative examples
  doubles negScores;
  int numNeg = 0;
  if (negSeqFile) {
    FILE *file = fopen(negSeqFile, "r");
    if (!file)
      error1("can't open %s", negSeqFile);

    Seq seq;
    while (getNextSeq(file, seq)) {
      BestSite bs;
      int locMax = seq.len - len;
      for (int loc=0; loc<=locMax; loc++) {
	bs(0, seq.seq, loc, true , len, (*ss)(seq.seq+loc,true ), 0);
	bs(0, seq.seq, loc, false, len, (*ss)(seq.seq+loc,false), 0);
      }
      negScores.push_back(bs.score);
    }
    numNeg = negScores.size();
    if (numNeg <= 1)
      error("too few negative examples");
    qsort(&negScores[0], numNeg, sizeof(double), _doubleRCmp);
  }

  // read sequence
  FILE *file = seqFile ? fopen(seqFile, "r") : stdin;
  if (!file) 
    error1("can't open %s", seqFile);

  Seq seq;
  while (getNextSeq(file, seq)) {
    // two types of mathematical criteria:
    //   score >= mc
    //   score >  mc2 
    double mc  = -DBL_MAX;
    double mc2 = -DBL_MAX;
    if (cutoff.use) {
      double c = cutoff.get(seq.desc);
      mc = max(mc, c);
    }
    if (pvalue.use) {
      double p = pvalue.get(seq.desc);
      double c = pvalueCutoff(p);
      mc = max(mc, c);
    }
    if (rank.use) {
      int r = (int)round(rank.get(seq.desc));
      if ((r < 1) || (r > numNeg))
	error1("rank out of range: %d", r);
      double c = negScores[r-1];
      mc2 = max(mc2, c);
    }
    if (fp.use) {
      double f = fp.get(seq.desc);
      if ((f < 0) || (f >= 1))
	error1("fp out of range: %f", f);
      // linearly interpolate
      // (i/(n-1), score[i])
      int i = (int)floor(f * (numNeg-1));
      int j = (int)ceil (f * (numNeg-1));
      double c = (i == j)
	? negScores[i]
	: linPred(
	    (double)i/(numNeg-1), negScores[i],
	    (double)j/(numNeg-1), negScores[j], f);
      mc2 = max(mc2, c);
    }
    
    // score subsequences
    Sites    sites;
    int      t = (int)round(top.get(seq.desc));
    TopSites topSites(t);

    int locMax = seq.len - len;
    for (int loc=0; loc<=locMax; loc++) {
      BestSite bs;
      bs(0, seq.seq, loc, true , len, (*ss)(seq.seq+loc, true ));
      bs(0, seq.seq, loc, false, len, (*ss)(seq.seq+loc, false));
      if ((bs.score < mc) || (bs.score <= mc2))
	continue;

      bs.pvalue = scorePvalue(bs.score);
      if (negSeqFile)
	scoreRankFp(bs.score, negScores, bs.rank, bs.fp);
	
      if (top.use)
	topSites(bs);
      else
	sites.push_back(bs.newSite());
    }
    if (top.use)
      topSites.getAll(sites);
    
    // print sites in sorted order
    qsort(&sites[0], sites.size(), sizeof(SitePtr), _siteRCmp);
    for (int i=0; i<sites.size(); i++) {
      Site &s = *sites[i];
      printf(">%s %d..%d %s %s %g",
	seq.sid,
	s.loc+1, s.loc+s.len,
	s.pStrand ? "forward" : "complementary",
	doubleStr(s.score),
	s.pvalue);
      if (negSeqFile)
	printf(" %d %s", s.rank, doubleStr(s.fp));
      printf("\n%s\n", s.seq);
    }
    deleteVec(sites);
  }
  if (seqFile)
    fclose(file);
  return 0;
}


/**
 * Extreme value distribution.
 *
 * [Mount, Bioinformatics, 2001, chapter 3] has a discussion
 * of the use of EVD in evaluating alignment scores.
 *
 * Note that the parameters in gsl differ from the ones presented in
 * [Mount].  In gsl, EVD is defined as:
 *   f = a b exp(-(b exp(-a x)+a x))
 *   F =     exp( -b exp(-a x)     )
 *
 * The distribution is scaled by 1/a and shifted by log(b).  Setting
 * both a and b to 1 gives
 *   f = exp(-exp(-x)-x)
 *   F = exp(-exp(-x)  )
 *
 * EVD satisfies the "stability postulate" F(x)^n = F(x-log(n)).
 * Stated differently, the maximum of n EVD random variables is
 * distributed identically to the original distribution, only shifted
 * by log(n).  F is characterized as the only non-periodic
 * distribution that satisfied the relationship F(x)^n = F(x-b_n).
 * EVD is also the asymptotic distribution of the maximum of n
 * unbounded random variables, independent of the original
 * distribution.
 *                  [Gumbel, Statistics of Extremes, 1958, chapter 5]
 * 
 * We'll find the generating function in order to calculate the mean
 * and standard deviation.
 *   g(t) = E[exp(t x)]
 *        = Integrate[exp(t x) exp(-exp(-x)-x), x=-infinity..infinity]
 *        = Integrate[  y^(-t) exp(-y)        , y=        0..infinity]
 *        = Gamma(1-t)
 *
 * Using Mathematica, we get
 *   mean = EulerGamma
 *   sd   = Pi / sqrt(6)
 *
 * The original distribution (with parameters a and b) has moments
 *   mean = (EulerGamma + log(b)) / a 
 *   sd   = Pi / (sqrt(6) a)
 *
 * Solving for a and b gives
 *   a = Pi / (sqrt(6) sd)
 *   b = exp(-EulerGamma + mean a)
 *
 * If we would like to take the maximum of n random such variables,
 * then multiply b by n.
 *   F^n = exp(-b   exp(-a x))^n
 *       = exp(-b n exp(-a x))
 **/
const double EulerGamma = 0.5772156649015328606065120900824024310422;
double evdA=1, evdB=1;
void setEvdParams (double mean, double sd, int n) {
  evdA = M_PI / (sqrt(6) * sd);
  evdB = exp(-EulerGamma + mean * evdA) * n;
}
double pvalueCutoff (double p) {
  /**
   * P[X >= cutoff] = pvalue
   * cutoff = Qinv[pvalue]
   **/
  return gsl_cdf_gumbel1_Qinv(p,evdA,evdB);
}
double scorePvalue (double s) {
  // P[X >= score] = pvalue
  return gsl_cdf_gumbel1_Q(s,evdA,evdB);
}


/**
 * Calculate rank and false-positive rate.
 **/
void scoreRankFp (double s, doubles &scores, int &r, double &f) {
  int n = scores.size();
  if (s > scores[0]) {
    r = 1;
    f = 0;
    return;
  }
  if (s <= scores[n-1]) {
    r = n+1;
    f = 1;
    return;
  }
  // binary search invariants:
  //   s <= scores[a]
  //   s >  scores[b]
  int a = 0;
  int b = n-1;
  while (b-a > 1) {
    int c = (a+b)/2;
    if (s <= scores[c])
      a = c;
    else
      b = c;
  }
  r = b + 1;  // handle ties conservatively
  f = linPred(
        scores[a], (double)a/(n-1),
	scores[b], (double)b/(n-1), s);
}
