/**
 * Motif Search is free software, you can redistribute it and/or
 * modify it under the terms of the GNU General Public License, a copy
 * of which is provided under the name LICENSE.
 *
 * Motif Search is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Copyright 2004 Robert Osada
 * $Id: motifsearch.h,v 1.5 2004/07/29 05:20:38 rosada Exp $
 *
 * load a motif and search for it in a sequence.
 * report sites which pass the specified criteria.
 *
 * written for Mona Singh's computational biology lab at
 * Princeton University, 2004.
 **/
#include <ctype.h>
#include <getopt.h>
#include <gsl/gsl_cdf.h>
#include "Fasta.h"
#include "SiteScorerFactory.h"

const char License[] =
  "This is Motif Search, version 1.0.\n\n"

  "Motif Search is free software; you can redistribute it and/or modify\n"
  "it under the terms of the GNU General Public License as published by\n"
  "the Free Software Foundation; either version 2 of the License, or\n"
  "(at your option) any later version.\n\n"
  
  "Motif Search is distributed in the hope that it will be useful,\n"
  "but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
  "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
  "GNU General Public License for more details.\n\n"
  
  "You should have received a copy of the GNU General Public License\n"
  "along with Motif Search; if not, write to the Free Software\n"
  "Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA\n\n"
  
  "Copyright 2004 Robert Osada\n"
  "Contact Mona Singh (msingh@cs.princeton.edu) for more information\n";

const char Usage[] =
  "usage:\n"
  "  motifsearch [options/method/criteria] motif seq > output\n\n"
  
  "options:                                                               default\n"
  "  -g, --gc content   gc content of random sequences                         .5\n"
  "  -L, --license      print out software license and exit\n\n"
  
  "method:\n"
  "  -m, --method string       set scoring method                        centroid\n"
  "               centroid     average match count\n"
  "               pssm         position specific scoring matrix\n"
  "               bergHippel   Berg & von Hippel method\n"
  "               consensus    consensus sequence\n"
  "  -i, --ic                  use information content                      false\n"
  "  -s, --scope  scope        set scope                                        0\n"
  "  -v, --var    variant      score variant\n\n"
  
  "criteria:                                                                top 1\n"
  "  -t, --top    n*        report top n scoring sites (per sequence)\n"
  "  -c, --cutoff cutoff*   report sites with score >= cutoff\n"
  "  -p, --pvalue pvalue*   choose a cutoff so that P[score >= cutoff] = pvalue\n"
  "                         for a randomly generated sequence of the same\n"
  "                         length as the motif\n"
  "  -n, --neg    negSeq    specify negative examples.\n"
  "                         each sequence contributes a single negative\n"
  "                         example--its highest scoring site.\n"
  "  -r, --rank   n*        report sites that score higher than\n"
  "                         the nth negative example\n"
  "  -f, --fp     rate*     choose a cutoff with an acceptable false\n"
  "                         positive rate.  specifically, a cutoff so that\n"
  "                         P[neg >= cutoff] = rate\n\n"

  "  * for starred criteria, use '$1','$2',... to read the value from the\n"
  "    sequence description (at the specified position)\n";


/**
 * Criteria.
 **/
struct Criteria {
  bool   use;
  int    pos;
  double val;
  Criteria () : use(false), pos(0), val(0) {}
  void set (cstring arg) {
    use = true;
    if (1 == sscanf(arg, " $ %d", &pos)) {
      if (pos < 0)
	error1("invalid position: %d", pos);
    }
    else
      val = atof(arg);
  }
  double get (cstring desc) {
    if (pos) {
      cstring t = findToken(desc, pos-1);
      if (sscanf(t, "%lf", &val) != 1)
	error2("unable to read criteria from position %d in '%s'", pos, desc);
    }
    return val;
  }
  static cstring findToken (cstring s, int i) {
    s += strspn(s, " \t");     // skip initial white space
    while (*s && i--) {
      s += strcspn(s, " \t");  // skip token
      s += strspn (s, " \t");  // skip white space
    }
    return s;
  }
};


/**
 * Extreme value distribution
 **/
void   setEvdParams (double mean, double sd, int n);
double pvalueCutoff (double p);
double scorePvalue  (double s);


/**
 * Calculate rank and false-positive rate.
 **/
void scoreRankFp (double s, doubles &scores, int &r, double &f);
