/**
 * This program is free software, you can redistribute it and/or
 * modify it under the terms of the GNU General Public License, a copy
 * of which is provided under the name LICENSE.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Copyright 2004 Robert Osada
 * $Id: Fasta.cpp,v 1.35 2005/01/21 20:46:49 rosada Exp $
 **/
#include "Fasta.h"
#include "Misc.h"

bool getNextSeq (FILE *file, Seq &seq, bool findSites) {
  const int   MAX_LINE_SIZE = 2048;
  static char line[MAX_LINE_SIZE], tmp[MAX_LINE_SIZE];
  assert(file);
  seq.clear();

  if (!readLine(file, line, MAX_LINE_SIZE))
    return false;
  if (1 != sscanf(line, "> %[^\n]", tmp))
    error("description doesn't start with '>'");
  seq.desc = strdup(tmp);

  // extract sequence id
  if (1 != sscanf(seq.desc, " %[^ ]", tmp))
    error1("can't parse sid from %s", seq.desc);
  seq.sid = strdup(tmp);
  
  // extract protein id
  cstring p = strchr(seq.sid, ':');
  if (p) {
    int len = p - seq.sid;
    seq.pid = strNew(len);
    strncpy(seq.pid, seq.sid, len);
    seq.pid[len] = 0;
  }
  
  // read sequence  
  int maxlen = 1024;
  seq.seq = strNew(maxlen);

  int c = fpeek(file);
  while ((c != EOF) && (c != '>')) {
    if (!readLine(file, line, MAX_LINE_SIZE))
      error1("error reading sequence %s", seq.sid);
    int l = strlen(line);
    if (seq.len + l > maxlen) {
      strResize(seq.seq, maxlen *= 2);
      if (!seq.seq)
	error1("sequence %s too long", seq.sid);
    }
    strcpy(seq.seq + seq.len, line);
    seq.len += l;
    c = fpeek(file);
  }
  // strip ends
  if (seq.seq[seq.len-1] == '/') { seq.seq[seq.len-1]=0; seq.len--; }
  if (seq.seq[seq.len-1] == '/') { seq.seq[seq.len-1]=0; seq.len--; }
  strResize(seq.seq, seq.len);
  
  seq.clearSites();
  if (findSites)
    seq.findSites();

  return true;
}

int readSeqs (cstring fileName, Seqs &seqs, bool findSites) {
  assert(fileName);
  FILE *file = fopen(fileName, "r");
  if (!file)
    error1("can't open %s", fileName);

  SeqPtr s = new Seq;
  int count = 0;

  while (getNextSeq(file, *s, findSites)) {
    seqs.push_back(s);
    s = new Seq;
    count++;
  }
  delete s;
  fclose(file);

  return count;
}

int readProteins (cstring fileName, SeqsVec &prots, bool findSites) {
  Seqs seqs;
  readSeqs(fileName, seqs, findSites);

  if (!seqs.size())
    return 0;

  SeqsPtr prot = new Seqs;
  prots.push_back(prot);
  int count = 1;

  prot->push_back(seqs[0]);
  cstring curPid = seqs[0]->pid;

  for (int i=1; i<seqs.size(); i++) {
    SeqPtr s = seqs[i];

    if (strcmp(curPid, s->pid)) {
      prot = new Seqs;
      prots.push_back(prot);
      count++;
    }
    prot->push_back(s);
    curPid = s->pid;
  }
  return count;
}

bool writeSeq (FILE *file, cstring desc, cstring seq, int len, int wrap) {
  assert(file && seq && (len >= 0) && (wrap > 0));

  if (desc)
    fprintf(file, ">%s\n", desc);

  int left = len;
  while (left) {
    int count = min(left, wrap);
    int res   = fwrite(seq, 1, count, file);
    if (res != count)
      return false;
    seq  += count;
    left -= count;
    fputc('\n', file);
  }
  return true;
}
bool writeSeq (FILE *file, Seq &seq, int wrap) {
  return writeSeq(file, seq.desc, seq.seq, seq.len, wrap);
}
bool writeSeq (FILE *file, Site &site, int wrap) {
  return writeSeq(file, site.desc, site.seq, site.len, wrap);
}
